! (C) 2008 D. V. Wiebe
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! This file is part of the GetData project.
!
! This program is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by the
! Free Software Foundation; either version 2 of the License, or (at your
! option) any later version.
!
! GetData is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
!
! You should have received a copy of the GNU General Public License along
! with GetData; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
!
! -*- F95 -*-

module getdata
implicit none


! Error codes
integer, parameter :: GD_E_OK=0
integer, parameter :: GD_E_OPEN=1
integer, parameter :: GD_E_FORMAT=2
integer, parameter :: GD_E_TRUNC=3
integer, parameter :: GD_E_CREAT=4
integer, parameter :: GD_E_BAD_CODE=5
integer, parameter :: GD_E_BAD_TYPE=6
integer, parameter :: GD_E_RAW_IO=7
integer, parameter :: GD_E_OPEN_INCLUDE=8
integer, parameter :: GD_E_INTERNAL_ERROR=9
integer, parameter :: GD_E_EMPTY=10
integer, parameter :: GD_E_ALLOC=11
integer, parameter :: GD_E_RANGE=12
integer, parameter :: GD_E_OPEN_LINFILE=13
integer, parameter :: GD_E_RECURSE_LEVEL=14
integer, parameter :: GD_E_BAD_DIRFILE=15
integer, parameter :: GD_E_BAD_PUT_FIELD=16
integer, parameter :: GD_E_ACCMODE=17

! Open flags
integer, parameter :: GD_RDONLY=0
integer, parameter :: GD_RDWR=2
integer, parameter :: GD_CREAT=64
integer, parameter :: GD_EXCL=128
integer, parameter :: GD_TRUNC=512
integer, parameter :: GD_BIG_ENDIAN=65536
integer, parameter :: GD_LITTLE_ENDIAN=131072
integer, parameter :: GD_FORCE_ENDIAN=262144
integer, parameter :: GD_PEDANTIC=524288

! Field types
integer, parameter :: GD_NO_ENTRY=0
integer, parameter :: GD_RAW_ENTRY=1
integer, parameter :: GD_LINCOM_ENTRY=2
integer, parameter :: GD_LINTERP_ENTRY=3
integer, parameter :: GD_BIT_ENTRY=4
integer, parameter :: GD_MULTIPLY_ENTRY=5
integer, parameter :: GD_PHASE_ENTRY=6

! Data types -- the unsigned type won't work when passed as a return
!               type, but we keep them anyways, since they might
!               appear as a result of calling fget_entry
integer, parameter :: GD_NULL=0
integer, parameter :: GD_UINT8=1
integer, parameter :: GD_INT8=17
integer, parameter :: GD_UINT16=2
integer, parameter :: GD_INT16=18
integer, parameter :: GD_UINT32=4
integer, parameter :: GD_INT32=20
integer, parameter :: GD_UINT64=8
integer, parameter :: GD_INT64=24
integer, parameter :: GD_FLOAT32=36
integer, parameter :: GD_FLOAT64=40


integer, parameter :: GD_FIELD_LEN=4096

type gd_entry
  integer :: field_type, n_fields, spf, data_type, bitnum, numbits, shift
  character(len=GD_FIELD_LEN), dimension(3) :: field
  real*8, dimension(3) :: m, b
end type

contains
function fdirfile_open (dirfilename, flags)
  integer :: fdirfile_open
  character (len=*), intent (in) :: dirfilename
  integer, intent (in) :: flags
  integer :: dirfile

  ! call f77 library
  call gdfopn(dirfile, TRIM(dirfilename), LEN_TRIM(dirfilename), flags)
  fdirfile_open = dirfile
end function

subroutine fdirfile_close (dirfile)
  integer, intent(in) :: dirfile

  ! call f77 library
  call gdfcls(dirfile)
end subroutine

subroutine fdirfile_flush (dirfile, field_code)
  integer, intent(in) :: dirfile
  character (len=*), intent(in) :: field_code

  ! call f77 library -- this will take care of figuring out whether
  !                     field_code is empty or not
  call gdffls(dirfile, TRIM(field_code), LEN_TRIM(field_code))
end subroutine

! getdata with return_type=GD_NULL
function fgetdata_n (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples)
  integer :: fgetdata_n
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_NULL, 0)
  fgetdata_n = nread
end function

! getdata with return_type=GD_INT8
function fgetdata_i1 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_i1
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*1, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT8, data_out)
  fgetdata_i1 = nread
end function

! getdata with return_type=GD_INT16
function fgetdata_i2 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_i2
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*2, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT16, data_out)
  fgetdata_i2 = nread
end function

! getdata with return_type=GD_INT32
function fgetdata_i4 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_i4
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*4, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT32, data_out)
  fgetdata_i4 = nread
end function

! getdata with return_type=GD_INT64
function fgetdata_i8 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_i8
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*8, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT64, data_out)
  fgetdata_i8 = nread
end function

! getdata with return_type=GD_FLOAT32
function fgetdata_r4 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_r4
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  real*4, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_FLOAT32, data_out)
  fgetdata_r4 = nread
end function

! getdata with return_type=GD_FLOAT64
function fgetdata_r8 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_out)
  integer :: fgetdata_r8
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  real*8, dimension(:), intent(out) :: data_out
  integer :: nread

  ! call f77 library
  call gdfget(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_FLOAT64, data_out)
  fgetdata_r8 = nread
end function

function fget_nfields (dirfile)
  integer :: fget_nfields
  integer, intent(in) :: dirfile
  integer :: nf

  ! call f77 library
  call gdfnfd(nf, dirfile)
  fget_nfields = nf
end function

function fget_field_name_max (dirfile)
  integer :: fget_field_name_max
  integer, intent(in) :: dirfile
  integer :: max_len

  ! call f77 library
  call gdffnx(max_len, dirfile)
  fget_field_name_max = max_len
end function

subroutine fget_field_list (field_list, dirfile, field_len)
  character(len=*), dimension(:), intent(out) :: field_list
  integer, intent(in) :: dirfile
  integer, intent(inout) :: field_len
  integer :: max_len, nfields, i

  ! make sure the field list is large enough
  max_len = fget_field_name_max(dirfile)

  if (field_len .lt. max_len) then
    field_len = max_len
  else
    nfields = fget_nfields(dirfile)
    do i=1,nfields
      ! call f77 library
      call gdffdn(field_list(i), field_len, dirfile, i)
    end do
  end if
end subroutine

function fget_nframes (dirfile)
  integer :: fget_nframes
  integer, intent(in) :: dirfile
  integer :: nf

  ! call f77 library
  call gdfnfr(nf, dirfile)
  fget_nframes = nf
end function

function fget_spf (dirfile, field_code)
  integer :: fget_spf
  integer, intent(in) :: dirfile
  character (len=*), intent(in) :: field_code
  integer :: spf

  ! call f77 library
  call gdfspf(spf, dirfile, TRIM(field_code), LEN_TRIM(field_code))
  fget_spf = spf
end function

! putdata with data_type=GD_INT8
function fputdata_i1 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_i1
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*1, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT8, data_in)
  fputdata_i1 = nread
end function

! putdata with data_type=GD_INT16
function fputdata_i2 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_i2
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*2, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT16, data_in)
  fputdata_i2 = nread
end function

! putdata with data_type=GD_INT32
function fputdata_i4 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_i4
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*4, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT32, data_in)
  fputdata_i4 = nread
end function

! putdata with data_type=GD_INT64
function fputdata_i8 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_i8
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  integer*8, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_INT64, data_in)
  fputdata_i8 = nread
end function

! putdata with data_type=GD_FLOAT32
function fputdata_r4 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_r4
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  real*4, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_FLOAT32, data_in)
  fputdata_r4 = nread
end function

! putdata with data_type=GD_FLOAT64
function fputdata_r8 (dirfile, field_code, first_frame, first_sample, &
  num_frames, num_samples, data_in)
  integer :: fputdata_r8
  integer, intent(in) :: dirfile, first_frame, first_sample, num_frames
  character (len=*), intent(in) :: field_code
  integer, intent(in) :: num_samples
  real*8, dimension(:), intent(in) :: data_in
  integer :: nread

  ! call f77 library
  call gdfput(nread, dirfile, TRIM(field_code), LEN_TRIM(field_code), &
  first_frame, first_sample, num_frames, num_samples, GD_FLOAT64, data_in)
  fputdata_r8 = nread
end function

function fget_error (dirfile)
  integer :: fget_error
  integer, intent(in) :: dirfile
  integer :: err

  ! call f77 library
  call gdferr(err, dirfile)
  fget_error = err
end function

subroutine fget_error_string (dirfile, buffer, len)
  integer, intent(in) :: dirfile, len
  character (len=*), intent(out) :: buffer

  ! call f77 library
  call gdfstr(dirfile, buffer, len)
end subroutine

function fget_entry (dirfile, field_code, ent)
  integer :: fget_entry
  integer, intent(in) :: dirfile
  character (len=*), intent(in) :: field_code
  type(gd_entry), intent(out) :: ent

  ! the call may modify the lengths, so don't use parameters
  integer :: len1 = GD_FIELD_LEN
  integer :: len2 = GD_FIELD_LEN
  integer :: len3 = GD_FIELD_LEN

  ! get field type from f77 library
  call gdffdt (fget_entry, dirfile, TRIM(field_code), LEN_TRIM(field_code))

  if (fget_entry .EQ. GD_RAW_ENTRY) then
    ! raw
    call gdferw(ent%spf, ent%data_type, dirfile, TRIM(field_code), &
    LEN_TRIM(field_code))
  else if (fget_entry .EQ. GD_LINCOM_ENTRY) then
    ! lincom
    call gdfelc(ent%n_fields, ent%field(1), len1, ent%m(1), ent%b(1), &
    ent%field(2), len2, ent%m(2), ent%b(2), ent%field(3), len3, ent%m(3), &
    ent%b(3), dirfile, TRIM(field_code), LEN_TRIM(field_code))
  else if (fget_entry .EQ. GD_LINTERP_ENTRY) then
    ! linterp
    call gdfelt(ent%field(1), len1, ent%field(2), len2, dirfile, &
    TRIM(field_code), LEN_TRIM(field_code))
  else if (fget_entry .EQ. GD_BIT_ENTRY) then
    ! bit
    call gdfebt(ent%field(1), len1, ent%bitnum, ent%numbits, dirfile, &
    TRIM(field_code), LEN_TRIM(field_code))
  else if (fget_entry .EQ. GD_MULTIPLY_ENTRY) then
    ! multiply
    call gdfemt(ent%field(1), len1, ent%field(2), len2, dirfile, &
    TRIM(field_code), LEN_TRIM(field_code))
  else if (fget_entry .EQ. GD_PHASE_ENTRY) then
    !phase
    call gdfemt(ent%field(1), len1, ent%shift, dirfile, TRIM(field_code), &
    LEN_TRIM(field_code))
  else
    ! on the off chance we got a type we weren't prepared to deal with
    fget_entry = GD_NO_ENTRY
  end if
end function

end module
